@@ -1,25 +1,34 @@
 #' Five times twofold cross-validation for the Hybrid Ensemble function
 #'
-#' \code{CVhybridEnsemble} cross-validates (five times twofold) the Hybrid Ensemble function and computes performance statistics that can be plotted (\code{\link{plot.CVhybridEnsemble}}) and summarized (\code{\link{summary.CVhybridEnsemble}}).
+#' \code{CVhybridEnsemble} cross-validates (five times twofold) (\code{\link{hybridEnsemble}}) and computes performance statistics that can be plotted (\code{\link{plot.CVhybridEnsemble}}) and summarized (\code{\link{summary.CVhybridEnsemble}}).
 #' 
 #' @param x A data frame of predictors. Categorical variables need to be transformed to binary (dummy) factors.
 #' @param y A factor of observed class labels (responses) with the only allowed values \{0,1\}.,
 #' @param combine Additional methods for combining the sub-ensembles. The simple mean, authority-based weighting and the single best are automatically provided since they are very effficient.  Possible additional methods: Genetic Algorithm: "rbga", Differential Evolutionary Algorithm: "DEopt", Generalized Simulated Annealing: "GenSA", Memetic Algorithm with Local Search Chains: "malschains", Particle Swarm Optimization: "psoptim", Self-Organising Migrating Algorithm: "soma", Tabu Search Algorithm: "tabu", Non-negative binomial likelihood: "NNloglik", Goldfarb-Idnani Non-negative least squares: "GINNLS", Lawson-Hanson Non-negative least squares: "LHNNLS".
-#' @param eval.measure Evaluation measure for the following combination methods: authority-based method, single best, "rbga", "DEopt","GenSA","malschains","psoptim","soma","tabu". Default is the area under the receiver operator characteristic curve 'auc'. The area under the sensitivity curve ('sens') and the area under the specificity curve ('spec') are also supported.
+#' @param eval.measure Evaluation measure for the following combination methods: authority-based method, single best, "rbga", "DEopt", "GenSA", "malschains", "psoptim", "soma", "tabu". Default is the area under the receiver operator characteristic curve 'auc'. The area under the sensitivity curve ('sens') and the area under the specificity curve ('spec') are also supported.
 #' @param verbose TRUE or FALSE. Should information be printed to the screen while estimating the Hybrid Ensemble.
+#' @param oversample TRUE or FALSE. Should oversampling be used? Setting oversample to TRUE helps avoid computational problems related to the subsetting process.
+#' @param filter either NULL (deactivate) or a percentage denoting the minimum class size of dummy predictors. This parameter is used to remove near constants. For example if nrow(xTRAIN)=100, and filter=0.01 then all dummy predictors with any class size equal to 1 will be removed. Set this higher (e.g., 0.05 or 0.10) in case of errors.
+#' @param LR.size Logistic Regression parameter. Ensemble size of the bagged logistic regression sub-ensemble.
 #' @param RF.ntree Random Forest parameter. Number of trees to grow.
 #' @param AB.iter Stochastic AdaBoost parameter. Number of boosting iterations to perform.
 #' @param AB.maxdepth Stochastic AdaBoost parameter. The maximum depth of any node of the final tree, with the root node counted as depth 0.
 #' @param KF.cp Kernel Factory parameter. The number of column partitions.
 #' @param KF.rp Kernel Factory parameter. The number of row partitions.
+#' @param KF.ntree Kernel Factory parameter. Number of trees to grow.
 #' @param NN.rang Neural Network parameter. Initial random weights on [-rang, rang].
 #' @param NN.maxit Neural Network parameter. Maximum number of iterations. 
-#' @param NN.size Neural Network parameter. Number of units in the single hidden layer.
-#' @param NN.decay Neural Network parameter. Weight decay.
-#' @param SV.gamma Support Vector Machines parameter. Width of the Guassian for radial basis and sigmoid kernel.
-#' @param SV.cost Support Vector Machines parameter. Penalty (soft margin constant).
-#' @param SV.degree Support Vector Machines parameter. Degree of the polynomial kernel.
-#' @param SV.kernel Support Vector Machines parameter. Kernels to try. Can be one or more of: 'radial','sigmoid','linear','polynomial'.
+#' @param NN.size Neural Network parameter. Number of units in the single hidden layer. Can be mutiple values that need to be optimized.
+#' @param NN.decay Neural Network parameter. Weight decay. Can be mutiple values that need to be optimized.
+#' @param NN.ens.size Neural Network parameter. Ensemble size of the neural network sub-ensemble.
+#' @param SV.gamma Support Vector Machines parameter. Width of the Guassian for radial basis and sigmoid kernel. Can be mutiple values that need to be optimized.
+#' @param SV.cost Support Vector Machines parameter. Penalty (soft margin constant). Can be mutiple values that need to be optimized.
+#' @param SV.degree Support Vector Machines parameter. Degree of the polynomial kernel. Can be mutiple values that need to be optimized.
+#' @param SV.kernel Support Vector Machines parameter. Kernels to try. Can be one or more of: 'radial','sigmoid','linear','polynomial'. Can be mutiple values that need to be optimized.
+#' @param SV.size Support Vector Machines parameter. Ensemble size of the SVM sub-ensemble.
+#' @param RoF.L Rotation Forest parameter. Number of trees to grow.
+#' @param KNN.K K-Nearest Neighbors parameter. Number of nearest neighbors to try. For example c(10,20,30). The optimal K will be selected. If larger than nrow(xTRAIN) the maximum K will be reset to 50\% of nrow(xTRAIN). Can be mutiple values that need to be optimized.
+#' @param KNN.size K-Nearest Neighbors parameter. Ensemble size of the K-nearest neighbor sub-ensemble.
 #' @param rbga.popSize Genetic Algorithm parameter. Population size.
 #' @param rbga.iters Genetic Algorithm parameter.  Number of iterations.
 #' @param rbga.mutationChance Genetic Algorithm parameter. The chance that a gene in the chromosome mutates.
@@ -87,26 +96,35 @@
 #' \item{SB}{For the single best: A list containing the median and inter quartile range of the performance evaluations, the performance evaluations on each fold, and the predictions and reponse vectors for each fold.}
 #' \item{eval.measure}{The performance measure that was used}
 #' ..and all the combination methods that are requested.
-#' @author Authors: Michel Ballings, Dauwe Vercamer, and Dirk Van den Poel, Maintainer: \email{Michel.Ballings@@GMail.com}
+#' @author Michel Ballings, Dauwe Vercamer, and Dirk Van den Poel, Maintainer: \email{Michel.Ballings@@GMail.com}
 
 CVhybridEnsemble <- function(x=NULL,
                              y=NULL,                             
                              combine=NULL,
                              eval.measure='auc', 
                              verbose=FALSE,
+                             oversample=TRUE,
+                             filter= 0.03,
+                             LR.size=10,
                              RF.ntree=500,
                              AB.iter=500,
                              AB.maxdepth=3,
                              KF.cp=1,
                              KF.rp=round(log(nrow(x),10)),
+                             KF.ntree=500,
                              NN.rang=0.1,
                              NN.maxit=10000,
                              NN.size=c(5,10,20),
                              NN.decay=c(0,0.001,0.01,0.1),
+                             NN.ens.size=10,
                              SV.gamma = 2^(-15:3),
                              SV.cost = 2^(-5:13),
                              SV.degree=c(2,3),
                              SV.kernel=c('radial','sigmoid','linear','polynomial'),
+                             SV.size=10,
+                             RoF.L=10,
+                             KNN.K=c(1:150),
+                             KNN.size=10,
                              rbga.popSize = 42,
                              rbga.iters = 500, 
                              rbga.mutationChance = 1/ rbga.popSize,
@@ -153,19 +171,28 @@ CVhybridEnsemble <- function(x=NULL,
                  combine=combine,
                  eval.measure=eval.measure,
                  verbose=FALSE,
+                 oversample=oversample,
+                 filter= filter,
+                 LR.size=LR.size,
                  RF.ntree=RF.ntree,
                  AB.iter=AB.iter,
                  AB.maxdepth=AB.maxdepth,
                  KF.cp=KF.cp,
                  KF.rp=KF.rp,
+                 KF.ntree=KF.ntree,
                  NN.rang=NN.rang,
                  NN.maxit=NN.maxit,
                  NN.size=NN.size,
                  NN.decay=NN.decay,
+                 NN.ens.size=NN.ens.size,
                  SV.gamma = SV.gamma,
                  SV.cost = SV.cost,
                  SV.degree=SV.degree,
                  SV.kernel=SV.kernel,
+                 SV.size=SV.size,
+                 RoF.L=RoF.L,
+                 KNN.K=KNN.K,
+                 KNN.size=KNN.size,
                  rbga.popSize = rbga.popSize,
                  rbga.iters = rbga.iters, 
                  rbga.mutationChance = rbga.mutationChance,
@@ -206,6 +233,20 @@ CVhybridEnsemble <- function(x=NULL,
                  tabu.iters=tabu.iters,
                  tabu.listSize=tabu.listSize
                  )
+  tab <- table(y)
+  if (!all(tab==tab[1])){
+      if (oversample) {
+
+        #oversample instances from the smallest class
+        whichmin <- which(y==as.integer(names(which.min(tab))))
+        indmin <- sample(whichmin,max(tab),replace=TRUE)
+        indmin <- c(whichmin,indmin)[1:max(tab)]
+        #take all the instances of the dominant class
+        indmax <- which(y==as.integer(names(which.max(tab))))
+        x <- x[c(indmin,indmax),]
+        y <- y[c(indmin,indmax)]    
+      }
+  }
   
   folds <- .partition(y,p=0.5,times=5)
   
@@ -279,7 +320,10 @@ CVhybridEnsemble <- function(x=NULL,
       xTEST <-  x[ folds[[i]][[iii]], ]
       yTEST <-  y[ folds[[i]][[iii]] ]
         
+
+      #remove constants and near constants
       constants <- sapply(xTRAIN,function(x){all(as.numeric(x[1])==as.numeric(x))})
+      if (!is.null(filter)) constants <- sapply(xTRAIN,function(x) length(unique(x))<=2 && any(table(x) <= round(nrow(xTRAIN)*filter)))
       xTRAIN <- xTRAIN[,!constants]
       xTEST <- xTEST[,!constants]
         
